from json import dumps, loads
from json.decoder import JSONDecodeError
from os import environ, path

from pytest import fixture, mark
from requests_oauth2client.tokens import BearerToken
from typer.testing import CliRunner

from sonyci import Config


def clean_response(response: dict):
    """Replace secrets in response body with dummy values."""
    try:
        body = loads(response['body']['string'].decode())
    except JSONDecodeError:
        return response
    if 'access_token' in body:
        body['access_token'] = 'DUMMY_ACCESS_TOKEN'
    if 'refresh_token' in body:
        body['refresh_token'] = 'DUMMY_REFRESH_TOKEN'
    response['body']['string'] = dumps(body).encode()
    return response


@fixture(scope='module')
def vcr_config(request):
    return {
        # Save cassettes in tests/vhs/<module_name>/<test_name>.yaml
        'cassette_library_dir': path.join(
            path.dirname(path.abspath(__file__)), 'vhs', request.module.__name__
        ),
        # Replace the Authorization request header with "DUMMY" in cassettes
        'filter_headers': [('authorization', 'Bearer DUMMY')],
        # Replace secrets in request body
        'filter_post_data_parameters': [
            ('client_id', 'FAKE_CLIENT_ID'),
            ('client_secret', 'FAKE_CLIENT_SECRET'),
        ],
        # Replace secrets in response body
        'before_record_response': clean_response,
    }


def pytest_addoption(parser):
    parser.addoption('--record', action='store_true', default=False)
    parser.addoption(
        '--no_ci',
        action='store_true',
        help='skips CI tests',
    )


def pytest_configure(config):
    config.addinivalue_line('markers', 'slow: mark test as slow to run')


def pytest_collection_modifyitems(config, items):
    if config.getoption('--no_ci'):
        skip_ci = mark.skip(reason='skipping CI tests')
        for item in items:
            if 'no_ci' in item.keywords:
                item.add_marker(skip_ci)


@fixture
def guid() -> str:
    return Config.from_toml('./tests/sonyci/guid.toml')['guid']


@fixture
def token() -> BearerToken:
    return dumps(
        BearerToken(
            access_token='FakeAccessToken', refresh_token='FakeRefreshToken'
        ).as_dict()
    )


# CLI fixtures
@fixture
def runner(token, pytestconfig):
    # If we're not recording, use a dummy token
    if not pytestconfig.getoption('record'):
        environ['CI_TOKEN'] = token
    return CliRunner()


@fixture
def error_runner():
    return CliRunner(mix_stderr=False)


@fixture
def config(pytestconfig):
    if pytestconfig.getoption('record'):
        return Config.from_toml('./ci.toml')
    return Config.from_toml('./tests/sonyci/sonyci.toml')


@fixture
def asset_id():
    return '554544ceaf6b4c94a4a06cee5bc1f39f'



from json import loads

from pytest import mark

from sonyci.cli import app


@mark.vcr()
def test_empty_search(runner, config):
    result = runner.invoke(
        app,
        [
            '-w',
            config['workspace_id'],
            'search',
            'i am not a guid',
        ],
    )
    assert result.exit_code == 0
    output = loads(result.output)
    assert isinstance(output, list)
    assert not output


@mark.vcr()
def test_guid_search(runner, config, guid):
    result = runner.invoke(
        app,
        [
            '-w',
            config['workspace_id'],
            'search',
            guid,
        ],
    )
    assert result.exit_code == 0
    output = loads(result.output)
    assert isinstance(output, list)
    assert len(output) == 1
    assert guid in output[0]['name']
    assert len(output) == 1
    assert guid in output[0]['name']



from pytest import fixture, mark

from sonyci import Config, SonyCi


@fixture(scope='module')
def ci_config(pytestconfig):
    if pytestconfig.getoption('record'):
        return Config.load('./ci.toml')
    return Config.load('./tests/sonyci/sonyci.toml')


@fixture(scope='module')
def ci(ci_config: Config):
    return SonyCi(**ci_config.model_dump())


@mark.vcr()
def test_token(ci: SonyCi):
    token = ci.token
    assert isinstance(token.access_token, str), 'Access token is not a string'
    assert len(token.access_token) > 0, 'Access token was empty'


@mark.vcr()
def test_workspaces(ci: SonyCi):
    workspaces = ci.workspaces()
    assert isinstance(workspaces, list), 'workspaces is not a list'
    assert len(workspaces) > 0, 'no workspaces found'


@mark.vcr()
def test_workspace(ci: SonyCi):
    workspace = ci.workspace()
    assert isinstance(workspace, dict), 'workspace is not a dict'
    assert 'id' in workspace, 'workspace has no id'


@mark.vcr()
def test_workspace_contents(ci: SonyCi):
    result = ci.workspace_contents()
    assert isinstance(result, list)
    assert result


@mark.vcr()
def test_workspace_empty_search(ci: SonyCi):
    result = ci.workspace_search(query='i am not a guid')
    assert isinstance(result, list)
    assert not result


@mark.vcr()
def test_workspace_search(ci: SonyCi, guid: str):
    assets = ci.workspace_search(guid)
    assert isinstance(assets, list)
    assert len(assets) == 1

    assert guid in assets[0]['name']


@mark.vcr()
def test_asset(ci: SonyCi, asset_id, **kwargs):
    asset = ci.asset(asset_id)
    assert isinstance(asset, dict)
    assert asset['id'] == asset_id


@mark.vcr()
def test_asset_download(ci: SonyCi, asset_id, **kwargs):
    asset = ci.asset_download(asset_id)
    assert isinstance(asset, dict)
    assert asset['id'] == asset_id



from typing import Any, ClassVar

from requests_oauth2client import ApiClient, OAuth2Client
from requests_oauth2client.auth import BearerAuth, OAuth2AccessTokenAuth
from requests_oauth2client.tokens import BearerToken

from sonyci.config import Config
from sonyci.log import log
from sonyci.utils import get_token, json


class SonyCi(Config):
    model_config: ClassVar = {'arbitrary_types_allowed': True}

    """A SonyCI client."""
    # This will not be needed when we upgrade to pydantic2,
    # we will be able to directly overwrite the @cached_property instance
    t: BearerToken = None

    @property
    def oauth(self) -> OAuth2Client:
        """Create and cache an OAuth2Client instance."""
        return OAuth2Client(
            token_endpoint=self.token_url,
            auth=(self.username, self.password),
            data={
                'grant_type': 'password',
                'client_id': self.client_id,
                'client_secret': self.client_secret,
            },
        )

    @property
    def token(self) -> BearerToken:
        """Get a token from SonyCI and cache the results."""
        if self.t:
            return self.t
        return get_token(
            username=self.username,
            password=self.password,
            client_id=self.client_id,
            client_secret=self.client_secret,
        )

    @property
    def auth(self) -> OAuth2AccessTokenAuth:
        """Create and cache an OAuth2AccessTokenAuth instance.

        This will refresh the token automatically if it is expired.
        """
        if self.client_id and self.client_secret:
            return OAuth2AccessTokenAuth(client=self.oauth, token=self.token)
        return BearerAuth(token=self.token)

    @property
    def client(self) -> ApiClient:
        """Create and cache an ApiClient instance.

        Example:
            ```py
            ci.client.get('workspaces')
            ```
        """
        return ApiClient(self.base_url, auth=self.auth)

    def workspaces(self) -> list:
        return self.get('workspaces')['items']

    def workspace(self) -> dict:
        """Return response of /workspaces/{workspace_id}"""
        return self.get(f'workspaces/{self.workspace_id}')

    def workspace_contents(self, **kwargs) -> list:
        """Returns items form the workspace"""
        return self.get(f'workspaces/{self.workspace_id}/contents', params=kwargs)[
            'items'
        ]

    def workspace_search(self, query: str | None = None, **kwargs) -> list:
        """Performs a search of a workspace and returns the items found"""
        if len(query) > 20:
            query = query[-20:]
        if query:
            kwargs['query'] = query
        if not kwargs.get('kind'):
            kwargs['kind'] = 'asset'
        return self.get(f'workspaces/{self.workspace_id}/search', params=kwargs)[
            'items'
        ]

    def asset(self, asset_id: str, **kwargs) -> dict:
        return self.get(f'/assets/{asset_id}', params=kwargs)

    def asset_download(self, asset_id: str, **kwargs) -> dict:
        return self.get(f'/assets/{asset_id}/download', params=kwargs)

    @json
    def get(self, *args, **kwargs):
        log.debug(f'GET {args} {kwargs}')
        return self.client.get(*args, **kwargs)

    @json
    def post(self, *args, **kwargs):
        return self.client.post(*args, **kwargs)

    def __call__(self, path: str, **kwds: Any) -> Any:
        """Default action: make GET request to Sony CI."""
        return self.get(path, **kwds)



from pytest import mark

from sonyci.cli import app


@mark.vcr()
def test_login(runner, config):
    result = runner.invoke(
        app,
        [
            '--client-id',
            config['client_id'],
            '--client-secret',
            config['client_secret'],
            'login',
            '--username',
            config['username'],
            '--password',
            config['password'],
            '--test',
        ],
    )
    assert result.exit_code == 0
    assert not result.stdout


@mark.vcr()
def test_bad_login(error_runner):
    result = error_runner.invoke(
        app,
        [
            '--client-id',
            'test',
            '--client-secret',
            'test',
            'login',
            '--username',
            'test',
            '--password',
            'test',
            '--test',
        ],
    )
    assert result.exit_code == 1
    assert 'invalid_client' in str(result.exception)


def test_missing_username(runner):
    result = runner.invoke(
        app,
        [
            '--client-id',
            'test',
            '--client-secret',
            'test',
            'login',
            '--password',
            'test',
        ],
    )
    assert result.exit_code == 2
    assert '--username' in result.stdout
